int utf8toucs2(int *utf8, int *ucs2);
int isutf8char(int ch);
int utoj(int *utf8, int *len);
int issjis(unsigned char ch);
int jtos(unsigned short ch);
int stoj(unsigned short ch);
int etoj(unsigned short ch);
int iseucchar(int ch);
int jtoe(unsigned short ch);
int iskana(unsigned char ch);
int validjis(unsigned short ch);

/* バイトの分解と合成 */
#define lobyte(x) (((unsigned short)(x)) & 0xff)
#define hibyte(x) ((((unsigned short)(x)) >> 8) & 0xff)
#define hilo(h,l) ((((unsigned char)(h)) << 8) | ((unsigned char)(l)))

